/*
 * This file is part of the AprilTag library.
 *
 * AprilTag is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * AprilTag is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zarray.h"

#define MIN_ALLOC 8

zarray_t *zarray_create(size_t el_sz)
{
    zarray_t *za = calloc(1, sizeof(zarray_t));
    za->el_sz = el_sz;
    return za;
}

void zarray_destroy(zarray_t *za)
{
    if (za == NULL)
        return;

    if (za->data != NULL)
        free(za->data);
    free(za);
}

inline void zarray_ensure_capacity(zarray_t *za, int capacity)
{
    if (za->alloc <= capacity) {

        capacity = za->alloc * 2;

        if (capacity < MIN_ALLOC)
            capacity = MIN_ALLOC;

        za->data = realloc(za->data, za->el_sz * capacity);
        za->alloc = capacity;
    }
}

// pass a pointer to the value you wish to store.
void zarray_add(zarray_t *za, const void *p)
{
    zarray_ensure_capacity(za, za->size + 1);

    memcpy(&za->data[za->size*za->el_sz], p, za->el_sz);
    za->size++;
}

// remove the entry whose value is equal to the value pointed to by p.
// if shuffle is true, the last element in the array will be placed in
// the newly-open space; if false, the zarray is compacted.
int zarray_remove_value(zarray_t *za, const void *p, int shuffle)
{
    for (int idx = 0; idx < za->size; idx++) {
        if (!memcmp(p, &za->data[idx*za->el_sz], za->el_sz)) {
            zarray_remove_index(za, idx, shuffle);
            return 1;
        }
    }

    return 0;
}

void zarray_remove_index(zarray_t *za, int idx, int shuffle)
{
    if (shuffle) {
        memcpy(&za->data[idx*za->el_sz], &za->data[(za->size-1)*za->el_sz], za->el_sz);
        za->size--;
        return;
    } else {
        // size = 10, idx = 7. Should copy 2 entries (at idx=8 and idx=9).
        // size = 10, idx = 9. Should copy 0 entries.
        int ncopy = za->size - idx - 1;
        if (ncopy > 0)
            memmove(&za->data[idx*za->el_sz], &za->data[(idx+1)*za->el_sz], ncopy*za->el_sz);
        za->size--;
        return;
    }
}

void zarray_insert(zarray_t *za, int idx, const void *p)
{
    zarray_ensure_capacity(za, za->size + 1);
    // size = 10, idx = 7. Should copy three entries (idx=7, idx=8, idx=9)
    int ncopy = za->size - idx;

    memmove(&za->data[(idx+1)*za->el_sz], &za->data[idx*za->el_sz], ncopy*za->el_sz);
    memcpy(&za->data[idx*za->el_sz], p, za->el_sz);

    za->size++;
}

void zarray_set(zarray_t *za, int idx, const void *p, void *outp)
{
    if (outp != NULL)
        memcpy(outp, &za->data[idx*za->el_sz], za->el_sz);

    memcpy(&za->data[idx*za->el_sz], p, za->el_sz);
}

void zarray_map(zarray_t *za, void (*f)())
{
    for (int idx = 0; idx < za->size; idx++)
        f(&za->data[idx*za->el_sz]);
}

void zarray_sort(zarray_t *za, int (*compar)(const void*, const void*))
{
    qsort(za->data, za->size, za->el_sz, compar);
}
