/*
 * This file is part of the AprilTag library.
 *
 * AprilTag is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * AprilTag is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#ifndef _ZARRAY_H
#define _ZARRAY_H

#include <stdlib.h>
#include <string.h>

typedef struct zarray zarray_t;
struct zarray
{
    size_t el_sz; // size of each element

    int size; // how many elements?
    int alloc; // we've allocated storage for how many elements?
    char *data;
};

zarray_t *zarray_create(size_t el_sz);
void zarray_destroy(zarray_t *za);

static inline void zarray_clear(zarray_t *za)
{
    za->size = 0;
}

static inline int zarray_size(zarray_t *za)
{
    return za->size;
}

void zarray_ensure_capacity(zarray_t *za, int capacity);

// pass a pointer to the value you wish to store.
void zarray_add(zarray_t *za, const void *p);

// pass a pointer to storage for the value you wish to retrieve
static inline void zarray_get(zarray_t *za, int idx, void *p)
{
    memcpy(p, &za->data[idx*za->el_sz], za->el_sz);
}

// similar to zarray_get, but returns a "live" pointer to the internal
// storage, avoiding a memcpy. This pointer is not valid across
// operations which might move memory around.
static inline void zarray_get_volatile(zarray_t *za, int idx, void *p)
{
    *((void**) p) = &za->data[idx*za->el_sz];
}

// remove the entry whose value is equal to the value pointed to by p.
// if shuffle is true, the last element in the array will be placed in
// the newly-open space; if false, the zarray is compacted. At most
// one element will be removed.
//
// Note that objects will be compared using memcmp over the full size
// of the value. If the value is a struct that contains padding,
// differences in the padding bytes can cause comparisons to
// fail. Thus, it remains best practice to bzero all structs so that
// the padding is set to zero.
//
// Returns the number of elements removed (0 or 1).
int zarray_remove_value(zarray_t *za, const void *p, int shuffle);

void zarray_remove_index(zarray_t *za, int idx, int shuffle);

void zarray_insert(zarray_t *za, int idx, const void *p);

void zarray_set(zarray_t *za, int idx, const void *p, void *outp);

void zarray_map(zarray_t *za, void (*f)());

// will call qsort internally. Your compare function will be called
// with pointers to the structures.
void zarray_sort(zarray_t *za, int (*compar)(const void*, const void*));

#endif
