/*
 * This file is part of the AprilTag library.
 *
 * AprilTag is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * AprilTag is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "zhash.h"

#define INITIAL_NBUCKETS 16
#define INITIAL_BUCKET_CAPACITY 4

// when the ratio of allocations to actual size drops below this
// ratio, we rehash. (Reciprocal of more typical load factor.)
#define REHASH_RATIO 2

struct bucket
{
    uint32_t size; // # entries in this bucket
    uint32_t alloc; // # entries allocated

    uint8_t *keys;
    uint8_t *values;
};

struct zhash
{
    size_t keysz, valuesz;

    uint32_t(*hash)(const void *a);

    // returns 1 if equal
    int(*equals)(const void *a, const void *b);

    int size; // # of items in hash table

    struct bucket *buckets;
    int nbuckets;
};

zhash_t *zhash_create(size_t keysz, size_t valuesz,
                      uint32_t(*hash)(const void *a), int(*equals)(const void *a, const void*b))
{
    zhash_t *vh = (zhash_t*) calloc(1, sizeof(zhash_t));
    vh->keysz = keysz;
    vh->valuesz = valuesz;
    vh->hash = hash;
    vh->equals = equals;

    vh->nbuckets = INITIAL_NBUCKETS;
    vh->buckets = (struct bucket*) calloc(vh->nbuckets, sizeof(struct bucket));
    return vh;
}

void zhash_destroy(zhash_t *vh)
{
    if (vh == NULL)
        return;

    for (int i = 0; i < vh->nbuckets; i++) {
        free(vh->buckets[i].keys);
        free(vh->buckets[i].values);
    }

    free(vh->buckets);
    free(vh);
}

int zhash_size(zhash_t *vh)
{
    return vh->size;
}

int zhash_contains(zhash_t *vh, const void *key)
{
    return zhash_get(vh, key, NULL);
}

int zhash_get(zhash_t *vh, const void *key, void *out_value)
{
    uint32_t hash = vh->hash(key);
    int idx = hash % vh->nbuckets;

    struct bucket *bucket = &vh->buckets[idx];
    for (int i = 0; i < bucket->size; i++) {
        void *this_key = &bucket->keys[vh->keysz * i];

        if (vh->equals(key, this_key)) {
            void *this_value = &bucket->values[vh->valuesz * i];
            if (out_value != NULL)
                memcpy(out_value, this_value, vh->valuesz);
            return 1;
        }
    }

    return 0;
}

int zhash_get_volatile(zhash_t *vh, const void *key, void *out_value)
{
    uint32_t hash = vh->hash(key);
    int idx = hash % vh->nbuckets;

    struct bucket *bucket = &vh->buckets[idx];
    for (int i = 0; i < bucket->size; i++) {
        void *this_key = &bucket->keys[vh->keysz * i];

        if (vh->equals(key, this_key)) {
            void *this_value = &bucket->values[vh->valuesz * i];
            if (out_value != NULL)
                *((void**) out_value) = this_value;
            return 1;
        }
    }

    return 0;
}

// returns 1 if there was an oldkey/oldvalue.
static inline int zhash_put_real(zhash_t *vh, struct bucket *buckets, int nbuckets,
                                 const void *key, const void *value, void *old_key, void *old_value)
{
    uint32_t hash = vh->hash(key);

    int idx = hash % nbuckets;
    struct bucket *bucket = &buckets[idx];

    // replace an existing key if it exists.
    for (int i = 0; i < bucket->size; i++) {
        void *this_key = &bucket->keys[vh->keysz * i];
        if (vh->equals(key, this_key)) {
            void *this_value = &bucket->values[vh->valuesz * i];

            if (old_key)
                memcpy(old_key, this_key, vh->keysz);

            if (old_value)
                memcpy(old_value, this_value, vh->valuesz);

            memcpy(this_key, key, vh->keysz);
            memcpy(this_value, value, vh->valuesz);
            return 1;
        }
    }

    // enlarge bucket?
    if (bucket->size == bucket->alloc) {
        if (bucket->alloc == 0)
            bucket->alloc = INITIAL_BUCKET_CAPACITY;
        else
            bucket->alloc *= 2;

        bucket->keys = realloc(bucket->keys, bucket->alloc * vh->keysz);
        bucket->values = realloc(bucket->values, bucket->alloc * vh->valuesz);
    }

    // add!
    void *this_key = &bucket->keys[vh->keysz * bucket->size];
    void *this_value = &bucket->values[vh->valuesz * bucket->size];

    memcpy(this_key, key, vh->keysz);
    memcpy(this_value, value, vh->valuesz);
    bucket->size++;

    if (old_key)
        memset(old_key, 0, vh->keysz);
    if (old_value)
        memset(old_value, 0, vh->valuesz);

    return 0;
}

int zhash_remove(zhash_t *vh, const void *key, void *old_key, void *old_value)
{
    uint32_t hash = vh->hash(key);

    int idx = hash % vh->nbuckets;
    struct bucket *bucket = &vh->buckets[idx];

    // replace an existing key if it exists.
    for (int i = 0; i < bucket->size; i++) {
        void *this_key = &bucket->keys[vh->keysz * i];

        if (vh->equals(key, this_key)) {
            if (old_key)
                memcpy(old_key, this_key, vh->keysz);

            void *this_value = &bucket->values[vh->valuesz * i];
            if (old_value)
                memcpy(old_value, this_value, vh->valuesz);

            // If we're removing the last element in the list,
            // then the shuffle remove would rewrite the last element
            // with itself. (That's forbidden with memcpy, so
            // explicitly test for it.)
            if (i != bucket->size - 1) {
                // shuffle remove.
                void *last_key = &bucket->keys[vh->keysz * (bucket->size - 1)];
                void *last_value = &bucket->values[vh->valuesz * (bucket->size - 1)];

                memcpy(this_key, last_key, vh->keysz);
                memcpy(this_value, last_value, vh->valuesz);
            }

            bucket->size--;
            vh->size--;

            return 1;
        }
    }

    if (old_key)
        memset(old_key, 0, vh->keysz);
    if (old_value)
        memset(old_value, 0, vh->valuesz);

    return 0;
}

int zhash_put(zhash_t *vh, const void *key, const void *value, void *oldkey, void *oldvalue)
{
    if (vh->nbuckets * REHASH_RATIO < vh->size) {

        // resize
        int new_nbuckets = vh->nbuckets*2;
        struct bucket *new_buckets = calloc(new_nbuckets, sizeof(struct bucket));

        // put all our existing elements into the new hash table
        zhash_iterator_t vit;
        zhash_iterator_init(vh, &vit);
        void *key, *value;
        while (zhash_iterator_next_volatile(&vit, &key, &value)) {
            zhash_put_real(vh, new_buckets, new_nbuckets, key, value, NULL, NULL);
        }

        // free the old elements
        for (int i = 0; i < vh->nbuckets; i++) {
            free(vh->buckets[i].keys);
            free(vh->buckets[i].values);
        }
        free(vh->buckets);

        // switch to the new elements
        vh->nbuckets = new_nbuckets;
        vh->buckets = new_buckets;
    }

    int has_oldkey = zhash_put_real(vh, vh->buckets, vh->nbuckets, key, value, oldkey, oldvalue);
    if (!has_oldkey)
        vh->size++;

    return has_oldkey;
}

void zhash_iterator_init(zhash_t *vh, zhash_iterator_t *vit)
{
    vit->vh = vh;
    vit->bucket = 0;
    vit->idx = 0;
}

int zhash_iterator_next_volatile(zhash_iterator_t *vit, void *outkey, void *outvalue)
{
    zhash_t *vh = vit->vh;

    while (vit->bucket < vh->nbuckets) {
        struct bucket *bucket = &vh->buckets[vit->bucket];

        if (vit->idx < bucket->size) {
            void *this_key = &bucket->keys[vh->keysz * vit->idx];
            void *this_value = &bucket->values[vh->valuesz * vit->idx];

            *((void**) outkey) = this_key;
            *((void**) outvalue) = this_value;
            vit->idx++;

            return 1;
        }
        vit->bucket++;
        vit->idx = 0;
    }

    return 0;
}

int zhash_iterator_next(zhash_iterator_t *vit, void *outkey, void *outvalue)
{
    void *outkeyp, *outvaluep;

    int res = zhash_iterator_next_volatile(vit, &outkeyp, &outvaluep);
    if (res) {
        if (outkey)
            memcpy(outkey, outkeyp, vit->vh->keysz);
        if (outvalue)
            memcpy(outvalue, outvaluep, vit->vh->valuesz);
    }
    return res;
}

void zhash_iterator_remove(zhash_iterator_t *vit)
{
    zhash_t *vh = vit->vh;

    struct bucket *bucket = &vh->buckets[vit->bucket];
    void *this_key = &bucket->keys[vh->keysz * (vit->idx - 1)];

    zhash_remove(vh, this_key, NULL, NULL);
    vit->idx--;
}

void zhash_map_keys(zhash_t * vh, void (*f)())
{
    zhash_iterator_t itr;
    zhash_iterator_init(vh, &itr);

    void *key, *value;

    while(zhash_iterator_next_volatile(&itr, &key, &value)) {
        f(key);
    }
}

void zhash_map_values(zhash_t * vh, void (*f)())
{
    zhash_iterator_t itr;
    zhash_iterator_init(vh, &itr);

    void *key, *value;
    while(zhash_iterator_next_volatile(&itr, &key, &value)) {
        f(value);
    }
}

zarray_t *zhash_keys(zhash_t *zh)
{
    zarray_t *za = zarray_create(zh->keysz);

    zhash_iterator_t itr;
    zhash_iterator_init(zh, &itr);

    void *key, *value;

    while(zhash_iterator_next_volatile(&itr, &key, &value)) {
        zarray_add(za, key);
    }

    return za;
}

zarray_t *zhash_values(zhash_t *zh)
{
    zarray_t *za = zarray_create(zh->valuesz);

    zhash_iterator_t itr;
    zhash_iterator_init(zh, &itr);

    void *key, *value;

    while(zhash_iterator_next_volatile(&itr, &key, &value)) {
        zarray_add(za, value);
    }

    return za;
}


uint32_t zhash_uint32_hash(const void *_a)
{
    uint32_t a = *((uint32_t*) _a);
    return a;
}

int zhash_uint32_equals(const void *_a, const void *_b)
{
    uint32_t a = *((uint32_t*) _a);
    uint32_t b = *((uint32_t*) _b);

    return a==b;
}

uint32_t zhash_uint64_hash(const void *_a)
{
    uint64_t a = *((uint64_t*) _a);
    return (uint32_t) (a ^ (a >> 32));
}

int zhash_uint64_equals(const void *_a, const void *_b)
{
    uint64_t a = *((uint64_t*) _a);
    uint64_t b = *((uint64_t*) _b);

    return a==b;
}

int zhash_str_equals(const void *_a, const void *_b)
{
    char *a = * (char**)_a;
    char *b = * (char**)_b;

    return !strcmp(a, b);
}

uint32_t zhash_str_hash(const void *_a)
{
    char *a = * (char**)_a;

    int64_t hash = 0;
    while (*a != 0) {
        hash += *a;
        hash = (hash << 7) + (hash >> 23);
        a++;
    }

    return hash;
}
