/*
 * This file is part of the AprilTag library.
 *
 * AprilTag is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * AprilTag is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Libav; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#ifndef _ZHASH_H
#define _ZHASH_H

#include <stdint.h>
#include "zarray.h"

#ifdef __cplusplus
extern "C" {
#endif


// XXX const-ness

/** A basic hash table capable of storing a variety of key/value
 * pairs, with customizable hash() and equals() functions.
 *
 * The hash table does nothing about memory management for the keys
 * and values-- no internal copying, no freeing, etc. This means that
 * key/values that are "put" into the hashtable must remain
 * valid. Keys, in particular, should not be modified.
 *
 **/

    typedef struct zhash zhash_t;

// The contents of the iterator should be considered private. However,
// since our usage model prefers stack-based allocation of iterators,
// we must publicly declare them.
    struct zhash_iterator
    {
        zhash_t *vh;

        // these point to the next bucket/bucket-entry to be examined.
        int bucket;
        int idx;
    };

    typedef struct zhash_iterator zhash_iterator_t;

    zhash_t *zhash_create(size_t keysz, size_t valuesz,
                          uint32_t(*hash)(const void *a),
                          int(*equals)(const void *a, const void *b));

    void zhash_destroy(zhash_t *vh);

// zhash_contains can tell the difference between a key being unmapped or mapped to NULL.
    int zhash_contains(zhash_t *vh, const void *key);

// returns a new array containing all the keys in this hashtable
// user is responsible for eventually calling 'varray_destroy(key_set);'
//    varray_t *zhash_keys(zhash_t *vh);
//    varray_t *zhash_values(zhash_t *vh);

// returns 0 if the key is unmapped
    int zhash_get(zhash_t *vh, const void *key, void *out_value);

// Similar to zhash_get, but more dangerous. Provides a pointer to the zhash's
// internal storage.  This can be used to make simple modifications to
// the underlying data while avoiding the memcpys associated with
// zhash_get and zhash_put. However, some zhash operations (that
// resize the underlying storage, in particular) render this pointer
// invalid. For maximum safety, call no other zhash functions for the
// period during which you intend to use the pointer.
    int zhash_get_volatile(zhash_t *vh, const void *key, void *out_p);

// The mapping key->value is established, replacing any previous
// mapping for key. The new key and new value will be stored in
// memory, potentially replacing and old key (even if it is "equal" to
// the new key) and value. The old key and old value are returned so
// that they can be deallocated if necessary. If there is no previous
// mapping for key, oldkey and oldvalue are set to NULL.
//
// oldkey/oldvalue may be NULL, in which case they are not returned.
//
// Returns 1 if a previous key/value mapping was found.
//
// example for a hashtable mapping char* to varray_t*.
//
// char *oldkey;
// varray_t *oldvalue;
//
// if (zhash_put(vh, newkey, newvalue, &oldkey, &oldvalue)) {  <-- note &'s
//    free(oldkey);
//    varray_map(free);
//    varray_destroy(oldvalue);
// }
    int zhash_put(zhash_t *vh, const void *key, const void *value, void *oldkey, void *oldvalue);

// do not modify the hash table while iterating over the elements,
// though you can call zhash_iterator_remove.
    void zhash_iterator_init(zhash_t *vh, zhash_iterator_t *vit);

    // provide a pointer to storage for outkey/outvalue. (Can be
    // NULL). Returns 0 if there are no more elements.
    int zhash_iterator_next(zhash_iterator_t *vit, void *outkey, void *outvalue);

    int zhash_iterator_next_volatile(zhash_iterator_t *vit, void *outkey, void *outvalue);

// remove the last returned key/value pair.
    void zhash_iterator_remove(zhash_iterator_t *vit);

// returns 1 if the item was removed. If oldkey/oldvalue are non-NULL,
// they are set to the removed key/value, or NULL if no item was removed.
    int zhash_remove(zhash_t *vh, const void *key, void *oldkey, void *oldvalue);

    int zhash_size(zhash_t *vh);

// apply a pair of functions to every [key,value] pair in the hash
// (e.g. free()) OK to set a function pointer to NULL if not
// needed. Note that the function will be passed a pointer to the data
// (which it should not modify).
    void zhash_map_keys(zhash_t *vh, void (*f)());
    void zhash_map_values(zhash_t *vh, void (*f)());

    zarray_t *zhash_keys(zhash_t *zh);
    zarray_t *zhash_values(zhash_t *zh);


    uint32_t zhash_uint32_hash(const void *a);
    int zhash_uint32_equals(const void *a, const void *b);

    uint32_t zhash_uint64_hash(const void *a);
    int zhash_uint64_equals(const void *_a, const void *_b);

uint32_t zhash_str_hash(const void *a);

/**
 * Defines a function to compare zhash values for string input data.
 * Can be used with zhash_create() for a key size of sizeof(char*).
 */
int zhash_str_equals(const void *a, const void *b);


#ifdef __cplusplus
}
#endif

#endif
